/*
  Pins.h Directly access GPIO ports efficiently
  Copyright (C) 2020  Thomas Almy

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

  Contact the author at tom@almy.us */
#ifndef _PINS_H
#define _PINS_H
#include <avr/io.h>

#if defined (__AVR_ATmega32U4__)
// Arduino Leonardo or Micro

#define digital_SS digital_17
#define digital_SCK digital_15
#define digital_MOSI digital_16
#define digital_MISO digital_14

/* The gcc compiler will compress reads and stores of these registers
 * to single (atomic) instructions because addresses are in range to do
 * so. It is not obvious that something like PORTB |= 1 << 3 will do
 * the same instead of a load, or, store sequence, however it does. 
 * In processors like ATmega2560 (Arduino Mega 2560)
 * not all the ports are in this good address range so operations will
 * only be atomic for ports A through G */

struct bits
{
  /* PINB at 0x03, DDRB at 0x04, PORTB at 0x05 */
  unsigned digital_17:1;  //SS and RXLED
  unsigned digital_15:1;  //SCK
  unsigned digital_16:1;  //MOSI
  unsigned digital_14:1;  //MISO
  unsigned digital_8:1;
  unsigned digital_9:1;
  unsigned digital_10:1;
  unsigned digital_11:1;
  
  /* Skip two bytes to get to Port C's register */
  unsigned :16;

  /* PINC at 0x06, DDRC at 0x07, PORTC at 0x08 */
  unsigned :6;
  unsigned digital_5:1;
  unsigned digital_13:1;

  /* Skip two bytes to get to port D's register. */
  unsigned :16;

  /* PIND at 0x09, DDRD at 0x0a, PORTD at 0x0b */
  unsigned digital_3:1;
  unsigned digital_2:1;
  unsigned digital_1:1;
  unsigned digital_0:1;
  unsigned digital_4:1;
  unsigned digital_30:1; // TXLED
  unsigned digital_6:1;
  unsigned :1;

  /* Skip two bytes to get to port E's register. */
  unsigned :16;

  /* PINE at 0x0c, DDRE at 0x0d, PORTE at 0x0e */
  unsigned :6;
  unsigned digital_7:1;
  unsigned :1;

  /* Skip two bytes to get to port F's register. */
  unsigned :16;

  /* PINF at 0x0f, DDRF at 0x10, PORTF at 0x11 */
  unsigned analog_5:1;  // Note backwards pin numbering compared to ADC multiplexer!
  unsigned analog_4:1;  // These have digital pin numbers as well!
  unsigned :1;
  unsigned :1;
  unsigned analog_3:1;
  unsigned analog_2:1;
  unsigned analog_1:1;
  unsigned analog_0:1;

};

/* These macros are used to access the different register bits.
   For example port.digital_8  gives PORTB bit 0*/

#define pin (*(volatile struct bits *)(&PINB))
#define ddr (*(/*volatile*/ struct bits *)(&DDRB))
#define port (*(volatile struct bits *)(&PORTB))

/* not defining ddr for volatile means that setting consecutive bits
 * can be done more efficiently, however in that case it won't be
 * atomic. This will never be a problem unless ddr is changed after
 * enabling interrupts. */

#else
#if defined (__AVR_ATmega4809__)
// Arduino Nano Every

#define digital_SS digital_8
#define digital_SCK digital_13
#define digital_MOSI digital_11
#define digital_MISO digital_12

/* The gcc compiler will compress reads and stores of these registers
 * to single (atomic) instructions because addresses are in range to do
 * so. It is not obvious that something like VPORTA.DIR |= 1 << 3 will do
 * the same instead of a load, or, store sequence, however it does. 
 */

struct bits
{
  /* VPORTA.DIR at 0x00, VPORTA.OUT at 0x01, VPORTA.IN at 0x02, VPORTA.INTFLAGS at 0x03 */
  unsigned digital_2:1;
  unsigned digital_7:1;
  unsigned analog_4:1;
  unsigned analog_5:1;
  unsigned :1;
  unsigned :1;
  unsigned :1;  
  unsigned :1;
  
  /* Skip three bytes to get to Port B's register */
  unsigned :16;
  unsigned :8;
  
  /* VPORTB.DIR at 0x04, VPORTB.OUT at 0x05, VPORTB.IN at 0x06, VPORTB.INTFLAGS at 0x07 */
  unsigned digital_9:1;
  unsigned digital_10:1;
  unsigned digital_5:1;
  unsigned :1;
  unsigned :1;
  unsigned :1;
  unsigned :1;  
  unsigned :1;
  
  /* Skip three bytes to get to Port C's register */
  unsigned :16;
  unsigned :8;

  /* VPORTC as above */
  unsigned :1;
  unsigned :1;
  unsigned :1;
  unsigned :1;
  unsigned digital_1:1;
  unsigned digital_0:1;
  unsigned digital_4:1;  
  unsigned :1;
  
  /* Skip three bytes to get to Port D's register */
  unsigned :16;
  unsigned :8;

  /* VPORTD as above */
  unsigned analog_3:1;
  unsigned analog_2:1;
  unsigned analog_1:1;
  unsigned analog_0:1;
  unsigned analog_6:1;
  unsigned analog_7:1;
  unsigned :1;  
  unsigned aref_pin:1;
  
  /* Skip three bytes to get to Port E's register */
  unsigned :16;
  unsigned :8;
  
  /* VPORTE as above */
  unsigned digital_11:1;
  unsigned digital_12:1;
  unsigned digital_13:1;
  unsigned digital_8:1;
  unsigned :1;
  unsigned :1;
  unsigned :1;  
  unsigned :1;
  
  /* Skip three bytes to get to Port F's register */
  unsigned :16;
  unsigned :8;
  
    /* VPORTF as above */
  unsigned :1;
  unsigned :1;
  unsigned :1;
  unsigned :1;
  unsigned digital_6:1;
  unsigned digital_3:1;
  unsigned reset_pin:1;  
  unsigned :1;
  };


/* These macros are used to access the different register bits.
   For example port.digital_8  gives VPORTB.OUT bit 0
   I'm using the same names for these as the other AVRs. */

#define ddr (*(/*volatile*/ struct bits *)(&VPORTA.DIR))
#define pin (*(volatile struct bits *)(&VPORTA.IN))
#define port (*(volatile struct bits *)(&VPORTA.OUT))

#define intflg (*(volatile struct bits *)(&VPORTA.INTFLAGS))


/* not defining ddr for volatile means that setting consecutive bits
 * can be done more efficiently, however in that case it won't be
 * atomic. This will never be a problem unless ddr is changed after
 * enabling interrupts. */

#else
#ifdef DDRE // Arduino Mega

#define digital_SS digital_53
#define digital_SCK digital_52
#define digital_MOSI digital_51
#define digital_MISO digital_50


/* The gcc compiler will compress reads and stores of these registers
 * to single (atomic) instructions because addresses are in range to do
 * so. It is not obvious that something like PORTB |= 1 << 3 will do
 * the same instead of a load, or, store sequence, however it does. 
 * In processors like ATmega2560 (Arduino Mega 2560)
 * not all the ports are in this good address range so operations will
 * only be atomic for ports A through G */

struct bits
{
  /* PINA at 0x00, DDRA at 0x01, PORTA at 0x02 */
  unsigned digital_22:1;
  unsigned digital_23:1;
  unsigned digital_24:1;
  unsigned digital_25:1;
  unsigned digital_26:1;
  unsigned digital_27:1;
  unsigned digital_28:1;
  unsigned digital_29:1; 
  
  /* Skip two bytes to get to Port B's register */
  unsigned :16;
  
  /* PINB at 0x03, DDRB at 0x04, PORTB at 0x05 */
  unsigned digital_53:1;
  unsigned digital_52:1;
  unsigned digital_51:1;
  unsigned digital_50:1;
  unsigned digital_10:1;
  unsigned digital_11:1;
  unsigned digital_12:1;
  unsigned digital_13:1;
  
  /* Skip two bytes to get to Port C's register */
  unsigned :16;

  /* PINC at 0x06, DDRC at 0x07, PORTC at 0x08 */
  unsigned digital_37:1;
  unsigned digital_36:1;
  unsigned digital_35:1;
  unsigned digital_34:1;
  unsigned digital_33:1;
  unsigned digital_32:1;
  unsigned digital_31:1;
  unsigned digital_30:1;

  /* Skip two bytes to get to port D's register. */
  unsigned :16;

  /* PIND at 0x09, DDRD at 0x0a, PORTD at 0x0b */
  unsigned digital_21:1;
  unsigned digital_20:1;
  unsigned digital_19:1;
  unsigned digital_18:1;
  unsigned :3; // Not accessible on Arduino board
  unsigned digital_38:1;

  /* Skip two bytes to get to Port E's register */
  unsigned :16;

  /* PINE at 0x0c, DDRE at 0x0d, PORTE at 0x0e */
  unsigned digital_0:1;
  unsigned digital_1:1;
  unsigned :1; // Not accessible on Arduino board
  unsigned digital_5:1;
  unsigned digital_2:1;
  unsigned digital_3:1;
  unsigned :2; // Not accessible on Arduino board

  /* Skip two bytes to get to Port F's register */
  unsigned :16;

  /* PINF at 0x0f, DDRF at 0x10, PORTF at 0x11 */
  unsigned analog_0:1;
  unsigned analog_1:1;
  unsigned analog_2:1;
  unsigned analog_3:1;
  unsigned analog_4:1;
  unsigned analog_5:1;
  unsigned analog_6:1;
  unsigned analog_7:1;

  /* Skip two bytes to get to Port G's register */
  unsigned :16;

  /* PING at 0x12, DDRG at 0x13, PORTG at 0x14 */
  unsigned digital_41:1;
  unsigned digital_40:1;
  unsigned digital_39:1;
  unsigned :2; // Not accessible on Arduino board
  unsigned digital_4:1;
  unsigned :2; // Not used
  
  /* Skip 205 bytes to get to port H's registers. */
  unsigned :8;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  unsigned :16;
  
  /* PINH at 0x100, DDRH at 0x101, PORTH at 0x102 */
  unsigned digital_17:1;
  unsigned digital_16:1;
  unsigned :1; // Not accessible on Arduino board
  unsigned digital_6:1;
  unsigned digital_7:1;
  unsigned digital_8:1;
  unsigned digital_9:1;
  unsigned :1; // Not accessible on Arduino board

  /* Skip two bytes to get to port J's register. */
  unsigned :16;
  /* PINJ at 0x103, DDRJ at 0x104, PORTJ at 0x105 */
  unsigned digital_15:1;
  unsigned digital_14:1;
  unsigned :6; // Not accessible on Arduino board

  /* Skip two bytes to get to port K's register. */
  unsigned :16;

  /* PINK at 0x106, DDRK at 0x107, PORTK at 0x108 */
  unsigned analog_8:1;
  unsigned analog_9:1;
  unsigned analog_10:1;
  unsigned analog_11:1;
  unsigned analog_12:1;
  unsigned analog_13:1;
  unsigned analog_14:1;
  unsigned analog_15:1;

  /* Skip two bytes to get to port L's register. */
  unsigned :16;

  /* PINL at 0x109, DDRL at 0x10a, PORTL at 0x10b */
  unsigned digital_49:1;
  unsigned digital_48:1;
  unsigned digital_47:1;
  unsigned digital_46:1;
  unsigned digital_45:1;
  unsigned digital_44:1;
  unsigned digital_43:1;
  unsigned digital_42:1;

};


/* These macros are used to access the different register bits.
   For example port.digital_8  gives PORTB bit 0*/

#define pin (*(volatile struct bits *)(&PINA))
#define ddr (*(/*volatile*/ struct bits *)(&DDRA))
#define port (*(volatile struct bits *)(&PORTA))

/* not defining ddr for volatile means that setting consecutive bits
 * can be done more efficiently, however in that case it won't be
 * atomic. This will never be a problem unless ddr is changed after
 * enabling interrupts. */

#else
// Arduino UNO

#define digital_SS digital_10
#define digital_SCK digital_13
#define digital_MOSI digital_11
#define digital_MISO digital_12

/* The gcc compiler will compress reads and stores of these registers
 * to single (atomic) instructions because addresses are in range to do
 * so. It is not obvious that something like PORTB |= 1 << 3 will do
 * the same instead of a load, or, store sequence, however it does. 
 * In processors like ATmega2560 (Arduino Mega 2560)
 * not all the ports are in this good address range so operations will
 * only be atomic for ports A through G */

struct bits
{
  /* PINB at 0x03, DDRB at 0x04, PORTB at 0x05 */
  unsigned digital_8:1;
  unsigned digital_9:1;
  unsigned digital_10:1;
  unsigned digital_11:1;
  unsigned digital_12:1;
  unsigned digital_13:1;
  unsigned :2;  /* PB6 and PB7 are not accessible and reserved for the crystal */
  
  /* Skip two bytes to get to Port C's register */
  unsigned :16;

  /* PINC at 0x06, DDRC at 0x07, PORTC at 0x08 */
  unsigned analog_0:1;
  unsigned analog_1:1;
  unsigned analog_2:1;
  unsigned analog_3:1;
  unsigned analog_4:1;
  unsigned analog_5:1;
  unsigned :2;  /* PC6 is not accessible (crystal) and PC7 is not used */

  /* Skip two bytes to get to port D's register. */
  unsigned :16;

  /* PIND at 0x09, DDRD at 0x0a, PORTD at 0x0b */
  unsigned digital_0:1;
  unsigned digital_1:1;
  unsigned digital_2:1;
  unsigned digital_3:1;
  unsigned digital_4:1;
  unsigned digital_5:1;
  unsigned digital_6:1;
  unsigned digital_7:1;

};


/* These macros are used to access the different register bits.
   For example port.digital_8  gives PORTB bit 0*/

#define pin (*(volatile struct bits *)(&PINB))
#define ddr (*(/*volatile*/ struct bits *)(&DDRB))
#define port (*(volatile struct bits *)(&PORTB))

/* not defining ddr for volatile means that setting consecutive bits
 * can be done more efficiently, however in that case it won't be
 * atomic. This will never be a problem unless ddr is changed after
 * enabling interrupts. */

#endif
#endif
#endif
#endif // _PINS_H
